# ifndef CPPAD_LOCAL_PLAY_ADDR_ENUM_HPP
# define CPPAD_LOCAL_PLAY_ADDR_ENUM_HPP
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
// SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
// SPDX-FileContributor: 2003-22 Bradley M. Bell
// ----------------------------------------------------------------------------

// BEGIN_CPPAD_LOCAL_PLAY_NAMESPACE
namespace CppAD { namespace local { namespace play {

/*!
\file addr_enum.hpp
*/
/// enum corresponding to type used for addressing iterators for a player
enum addr_enum {
   unsigned_short_enum  ,
   unsigned_int_enum    ,
   size_t_enum
};

} } } // BEGIN_CPPAD_LOCAL_PLAY_NAMESPACE

# endif
