# ifndef CPPAD_CORE_FUN_EVAL_HPP
# define CPPAD_CORE_FUN_EVAL_HPP
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
// SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
// SPDX-FileContributor: 2003-22 Bradley M. Bell
// ----------------------------------------------------------------------------

# include <cppad/core/new_dynamic.hpp>
# include <cppad/core/forward/forward.hpp>
# include <cppad/core/reverse.hpp>
# include <cppad/core/sparse.hpp>

# endif
