# ifndef CPPAD_LOCAL_OP_HPP
# define CPPAD_LOCAL_OP_HPP
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
// SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
// SPDX-FileContributor: 2003-22 Bradley M. Bell
// ----------------------------------------------------------------------------

// used by the sparse operators
# include <cppad/local/sparse/internal.hpp>

// operations
# include <cppad/core/std_math_11.hpp>
# include <cppad/local/op/abs_op.hpp>
# include <cppad/local/op/add_op.hpp>
# include <cppad/local/op/acos_op.hpp>
# include <cppad/local/op/acosh_op.hpp>
# include <cppad/local/op/asin_op.hpp>
# include <cppad/local/op/asinh_op.hpp>
# include <cppad/local/op/atan_op.hpp>
# include <cppad/local/op/atanh_op.hpp>
# include <cppad/local/op/comp_op.hpp>
# include <cppad/local/op/cond_op.hpp>
# include <cppad/local/op/cos_op.hpp>
# include <cppad/local/op/cosh_op.hpp>
# include <cppad/local/op/cskip_op.hpp>
# include <cppad/local/op/csum_op.hpp>
# include <cppad/local/op/discrete_op.hpp>
# include <cppad/local/op/div_op.hpp>
# include <cppad/local/op/erf_op.hpp>
# include <cppad/local/op/exp_op.hpp>
# include <cppad/local/op/expm1_op.hpp>
# include <cppad/local/op/load_op.hpp>
# include <cppad/local/op/log_op.hpp>
# include <cppad/local/op/log1p_op.hpp>
# include <cppad/local/op/mul_op.hpp>
# include <cppad/local/op/neg_op.hpp>
# include <cppad/local/op/parameter_op.hpp>
# include <cppad/local/op/pow_op.hpp>
# include <cppad/local/op/print_op.hpp>
# include <cppad/local/op/sign_op.hpp>
# include <cppad/local/op/sin_op.hpp>
# include <cppad/local/op/sinh_op.hpp>
# include <cppad/local/op/sqrt_op.hpp>
# include <cppad/local/op/sub_op.hpp>
# include <cppad/local/sparse/binary_op.hpp>
# include <cppad/local/sparse/unary_op.hpp>
# include <cppad/local/op/store_op.hpp>
# include <cppad/local/op/tan_op.hpp>
# include <cppad/local/op/tanh_op.hpp>
# include <cppad/local/op/zmul_op.hpp>


# endif
