# ifndef CPPAD_EXAMPLE_EIGEN_PLUGIN_HPP
# define CPPAD_EXAMPLE_EIGEN_PLUGIN_HPP
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
// SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
// SPDX-FileContributor: 2003-22 Bradley M. Bell
// ----------------------------------------------------------------------------
/*$
{xrst_begin eigen_plugin.hpp}

Source Code for eigen_plugin.hpp
################################
{xrst_spell_off}
{xrst_code cpp} */
// Declaration needed, before eigen-3.3.3, so Eigen vector is a simple vector
typedef Scalar value_type;
/* {xrst_code}
{xrst_spell_on}

{xrst_end eigen_plugin.hpp}
*/
# endif
