# ifndef CPPAD_EXAMPLE_MULTI_THREAD_TEAM_EXAMPLE_HPP
# define CPPAD_EXAMPLE_MULTI_THREAD_TEAM_EXAMPLE_HPP
// SPDX-License-Identifier: EPL-2.0 OR GPL-2.0-or-later
// SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
// SPDX-FileContributor: 2003-22 Bradley M. Bell
// ----------------------------------------------------------------------------

extern bool team_example(void);

# endif
